@@ -86,6 +86,9 @@
 # include <sys/ioctl.h>
 # include <ext2fs/ext2_fs.h>
 # include <utime.h>
+# ifdef HAVE_LIBMOUNT_LIBMOUNT_H
+#  include <libmount/libmount.h>
+# endif
 #endif
 
 /* do we need to work around a glibc dependency? */
@@ -300,6 +303,55 @@ safe_chdir (const char * dirname)
     return 0;
 }
 
+
+/*
+ * mountpoint stuff copied from mountpoint.c, part of util-linux and modified.
+ * Perhaps the mountpoint.c stuff could be used more extensively, replacing
+ * hand-coded stuff in tmpreaper
+ */
+#ifdef HAVE_LIBMOUNT_LIBMOUNT_H
+
+#define _PATH_PROC_MOUNTINFO	"/proc/self/mountinfo"
+
+struct mountpoint_control {
+	char *path;
+	dev_t dev;
+	struct stat st;
+	unsigned int
+		dev_devno:1,
+		fs_devno:1;
+};
+
+int
+dir_to_device(struct mountpoint_control *ctl)
+{
+	struct libmnt_table *tb = mnt_new_table_from_file(_PATH_PROC_MOUNTINFO);
+	struct libmnt_fs *fs;
+	struct libmnt_cache *cache;
+	int rc = -1;
+
+	if (!tb) {
+		/* Not possible to detect bind mounts */
+
+		return -1;
+	}
+
+	/* to canonicalize all necessary paths */
+	cache = mnt_new_cache();
+	mnt_table_set_cache(tb, cache);
+	mnt_unref_cache(cache);
+
+	fs = mnt_table_find_target(tb, ctl->path, MNT_ITER_BACKWARD);
+	if (fs && mnt_fs_get_target(fs)) {
+		ctl->dev = mnt_fs_get_devno(fs);
+		rc = 0;
+	}
+
+	mnt_unref_table(tb);
+	return rc;
+}
+#endif /* HAVE_LIBMOUNT_LIBMOUNT_H */
+
 int
 dir_empty_p (const char * dirname)
 {
@@ -332,7 +384,6 @@ cleanupDirectory (const char	   * dirname ,
 {
     DIR		  * dir;
     struct dirent * ent;
-    struct stat	    sb, here;
     int		    status, pid, skip = 0;
     int		    i;
     static int	    fork_count = 0;
@@ -404,6 +455,7 @@ cleanupDirectory (const char	   * dirname ,
      */
 
     if (! (pid = fork())) {
+	    struct stat	    sb, here;
 	    if (safe_chdir (dirname))
 		    exit(1);
 	    fork_count++;
@@ -467,28 +519,6 @@ cleanupDirectory (const char	   * dirname ,
 		    continue;
 		}
 
-		if (S_ISDIR (sb.st_mode)) {
-                    char *dst;
-
-                    if ((dst = malloc(strlen(ent->d_name) + 3)) == NULL)
-                        message (LOG_FATAL, "malloc failed.\n");
-                    strcpy(dst, ent->d_name);
-                    strcat(dst, "/X");
-                    rename(ent->d_name, dst);
-                    if (errno == EXDEV) {
-                        free(dst);
-                        message (LOG_VERBOSE,
-                                 "File on different device skipped: `%s/%s'\n",
-                                 dirname, ent->d_name);
-                        continue;
-                    }
-                    free(dst);
-		    cleanupDirectory (ent->d_name, killTime, flags);
-		    message (LOG_VERBOSE,
-			     "(PID %u) Back from recursing down `%s'.\n",
-			     (u_int) getpid(), ent->d_name);
-		}
-
 		if (FLAGS_PROTECT_P (flags)) {
 		    skip = i = 0;
 		    do {
@@ -504,6 +534,52 @@ cleanupDirectory (const char	   * dirname ,
 			continue;
 		}
 
+		if (S_ISDIR (sb.st_mode)) {
+#ifdef HAVE_LIBMOUNT_LIBMOUNT_H
+		    struct mountpoint_control ctl = { NULL };
+
+		    ctl.path = ent->d_name;
+		    if (stat(ctl.path, &ctl.st)) {
+			message (LOG_ERROR,
+				 "Failed to stat() `%s/%s/.': %s\n",
+				 dirname, ent->d_name, strerror (errno));
+			continue;
+		    }
+		    if (!dir_to_device(&ctl)) {
+			message (LOG_VERBOSE,
+				 "Directory on different device skipped: `%s/%s'\n",
+				 dirname, ent->d_name);
+			continue;
+		    }
+#else /* fallback for systems (non-Linux?) without libmount */
+                    char *dst;
+		    struct stat sb2;
+
+                    if ((dst = malloc(strlen(ent->d_name) + 3)) == NULL)
+                        message (LOG_FATAL, "malloc failed.\n");
+                    strcpy(dst, ent->d_name);
+                    strcat(dst, "/.");
+		    if (lstat (dst, &sb2)) {
+			message (LOG_ERROR,
+				 "Failed to lstat() `%s/%s/.': %s\n",
+				 dirname, ent->d_name, strerror (errno));
+			free(dst);
+			continue;
+		    }
+		    free(dst);
+		    if (sb2.st_dev != sb.st_dev) {
+			message (LOG_VERBOSE,
+				 "Directory on different device skipped: `%s/%s'\n",
+				 dirname, ent->d_name);
+			continue;
+		    }
+#endif
+		    cleanupDirectory (ent->d_name, killTime, flags);
+		    message (LOG_VERBOSE,
+			     "(PID %u) Back from recursing down `%s'.\n",
+			     (u_int) getpid(), ent->d_name);
+		}
+
 		/* Decide whether to remove the file or not */
 		/* check for mtime on directory instead of atime if requested */
 		if ( FLAGS_MTIME_P(flags) ||
@@ -649,7 +725,7 @@ printCopyright (void)
 	     "tmpreaper -- Version: " VERSION "\n"
 #endif
 	     "(c) 1997 Karl M. Hegbloom   <tmpreaper@packages.debian.org>\n"
-	     "(c) 2006-2008 Paul Slootman <tmpreaper@packages.debian.org>\n"
+	     "(c) 2006-2019 Paul Slootman <tmpreaper@packages.debian.org>\n"
 	     "This may be freely redistributed under the terms of the GNU Public License.\n");
 }
 /*	               1         2         3         4         5         6         7         8
@@ -799,7 +875,7 @@ main (int     argc,
         case 'T':
             runtimeMax = atoi(optarg);
             message(LOG_DEBUG, "runtime is %d seconds max\n", runtimeMax);
-            if (runtimeMax < 5) {
+            if (runtimeMax < 5 && runtimeMax != 0) {
                 message(LOG_FATAL, "runtime must be at least 5 seconds\n");
                 /*NOTREACHED*/
             }
@@ -921,8 +997,10 @@ main (int     argc,
         sleep(timetosleep);
     }
       
-    signal(SIGALRM, sig_alarmhandler);
-    alarm(runtimeMax);
+    if (runtimeMax) {
+        signal(SIGALRM, sig_alarmhandler);
+        alarm(runtimeMax);
+    }
 
     /* For each directory in the cleanup list, cd there, glob, then delete */
     for (adir = optind; adir < argc; adir++) {
@@ -978,7 +1056,8 @@ main (int     argc,
 		continue;
 	    }
 	    protect_table[j].inode = sb.st_ino;
-	    protect_table[j++].name = protect_glob.gl_pathv[j];
+	    protect_table[j].name = protect_glob.gl_pathv[j];
+	    j++;
 	}
 	protect_table[j].name = NULL;
 
